//**************************************************************
//*
//* Pinbelegung
//*    ----------------------------------
//*        1  VDD (+5V)
//*        2  GP5 -out(in)- (output (& gainselect in))
//*        3  GP4 -out- (discharge enable)
//*        4  GP3 -in - (start/reset)
//*        5  AN2 -in - (termination voltage)
//*        6  AN1 -in - (MAX REFOUT)
//*        7  AN0 -in - (MAX OUT)
//*        8  VSS (GND)
//*
//**************************************************************
//
// loetmeister.de - 2013, v0.8
//
// ADC for MAX4071:
//
// 12F675 wandelt die an AN0 - AN1 diff. anliegende Spannung in eine
// Pulsfolge um. Fr jede mAh ein Impuls. Ausgabe an GP4.
// Die Entlandung wird beendet wenn an AN2 die Abschaltspannung
// (bei mehreren aufeindander folgenden Messungen) unterschritten ist.
//
// http://loetmeister.de/Elektronik/akku_kapazitaetsmesser.htm
//
// Prozessortakt:  4 MHz intern
// Eingangsspannung: 5V

/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#include <xc.h>         /* XC8 General Include File */

#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */

#include "system.h"        /* System funct/params, like osc/peripheral config */
#include "user.h"          /* User funct/params, such as InitApp */

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/

signed int ADCresult=0; // calculated ADC result (Vinput - Vref)
int ADCcounter=0;   // counts up the ADC result
char ADCcounterTenths=const_ADCcounterTenths;    // use to correct the 0.5 result

char timer2=const_timer2;
char VcuttOffCounter=const_VcuttOffCounter;
char startButton_timer2=0;

bit runningState=0;         // status of dischaging
bit startButtonState=0;
bit startButtonStateOld=0;
bit gainSelect=0;           // gain factor of MAX4071 (low: 0=50mV/V, high: 1=100mV/V)

// to store the three ADC result values
union {unsigned int ADRES;
    char ad_hi_low[2];
} ADCref, ADCout, ADCVterm;

/******************************************************************************/
/* Main Program                                                               */
/******************************************************************************/
void main(void)
{
    /* Configure the oscillator for the device */
    //ConfigureOscillator(); // no need here, default startup code will do it

    /* Initialize I/O and Peripherals */
    InitApp();


    while(1)
    {
        /* set output pin to start or stop discharging
         * output = HIGH, starts discharging */
        outDischargePin = runningState;


        if (timer2 == const_timer2_half) {
            if  (runningState == 1) {
                // set output pin to LOW only if running and some time passed by...
                // (outPin was set when timer2 was zero and TMR0 interrupt occured)
                outPin = gainSelect; //rest output (active LOW for high gain operation and HIGH for low gain)
            }
            else {
                /* read the status of the gain select switch. We use the same pin
                 * also for the output, so only check port status when not in
                 * running mode. Set to correct output mode after reading! */
                outPinTRISIObit = 1; // set the port as input
                _delay(8);  // not sure if this is required, but give some micro seconds to settle
                gainSelect = outPin;    // read it
                outPinTRISIObit = 0; // set as output again
                outPin = gainSelect;    // write new state back, to avoid further output impuls
            }
        }

        /* moved button reading to normal code, not in interrupt routine */
        startButtonState = startButtonPin;  // read input
        
        // check if start button is still pressed
        if (startButtonState != startButtonStateOld) {
            startButton_timer2 = timer2; // reset startButton_timer2 if not
        }
        /* timer2 decrement by one at every TMR0 role over (62,5Hz) (so debounce is approx. 16ms)   */
        if (startButton_timer2 != timer2 && startButtonPin == 1) {  // we only want to know when button is pressed
            runningState = 1;
        }
        startButtonStateOld = startButtonState;
        
        
        // sample input for analog channel 1 (MAX vref)
        ADCON0bits.CHS = 0b01;    // select Analog Channel AN1

    /* min. wait time is 4s after channel select for A/D speed F OSC/16
     * to be on the save side, let's wait 8 s */
        _delay(8); // wait 8 cycles (1/(4MHz/4) = 1s * 8 = 8 s)
        adc_read();
        ADCref.ad_hi_low[0] = ADRESL;
        ADCref.ad_hi_low[1] = ADRESH;


        // sample input for analog channel 0 (MAX out)
        ADCON0bits.CHS = 0b00;    // select Analog Channel AN0

        _delay(8); // wait 8 cycles (1/(4MHz/4) = 1s * 8 = 8 s)
        adc_read();
        ADCout.ad_hi_low[0] = ADRESL;
        ADCout.ad_hi_low[1] = ADRESH;

        // calculate (absolute) result
        ADCresult = ADCout.ADRES - ADCref.ADRES;

        // sample input for analog channel 2 (termination voltage divider)
        ADCON0bits.CHS = 0b10;    // select Analog Channel AN2

        _delay(8); // wait 8 cycles (1/(4MHz/4) = 1s * 8 = 8 s)
        adc_read();
        ADCVterm.ad_hi_low[0] = ADRESL;
        ADCVterm.ad_hi_low[1] = ADRESH;


        /* check if termination voltage is smaller then the threshold */
        /* to run at timer2 frequency, only perform a check when timer2 has a specific value */
        if ((runningState == 1) && (timer2 == const_timer2_half)) {
            if (((ADCVterm.ADRES < VcuttOff_lowGain) && (gainSelect == 0)) || ((ADCVterm.ADRES < VcuttOff_highGain) && (gainSelect == 1))) {
                --VcuttOffCounter;
                if (VcuttOffCounter <= 0) {
                    runningState = 0;   // after n-times, stop discharging
                }
            }
            else {
                VcuttOffCounter = const_VcuttOffCounter;    // reset counter
            }
        }
        else {
            VcuttOffCounter = const_VcuttOffCounter;    // reset counter if not running
        }


    } // while
}

/******************************************************************************/
/* Interrupt routine                                                          */
/******************************************************************************/
void interrupt isr(void)
{
    // reset timer0 when interrupt has occured (62,5Hz)
    TMR0 = const_timer0;


    // check if timer2 ticked down
    if (timer2 == 0) {
        timer2 = const_timer2;  // reset timer2
        /* 2,0833Hz = (3750mA/3600s)*2 */
        /* 2,0833Hz = (2500mA/3600s)*3 */

        if (runningState == 1) {    // only if requested to work
            // add last result to counter
            ADCcounter += ADCresult;

            /* check if the result counter was reset x times */
            if (ADCcounterTenths <= 0 && gainSelect == 0) { // only for uneven counter value in low gain operation
                ADCcounter += 1;        // add one to result counter
                ADCcounterTenths = const_ADCcounterTenths;   // and reset
            }

            /* check if enough was "collected" for an output impuls
            * (small input values let this counter increase slower) */
            // low gain level
            if (ADCcounter >= maxVin_lowGain && gainSelect == 0) {
                outPin = 1; //Ausgang = HIGH

                // keep track of the tenth we "lost"
                --ADCcounterTenths;

                // keep rest of the result for the next round
                ADCcounter -= maxVin_lowGain;
            }
            // high gain level
            else if (ADCcounter >= maxVin_highGain && gainSelect == 1) {
               outPin = 0; //Ausgang = LOW (active low for high gain operation)

                // keep rest of the result for the next round
                ADCcounter -= maxVin_highGain;
            }
        }
        else {
            ADCcounter = 0; // reset counter if not in running state
            ADCcounterTenths = const_ADCcounterTenths;   // reset
        }
    }
    --timer2;

    INTCONbits.T0IF = 0;   // clear interrupt flag
}

/******************************************************************************/
// EOF